"""

Intro to project ...


Setup
=====

In order to use this library, ...


Main concepts
=============

Describe typical classes found in project

.. autoclass:: SampleClassName
   :members:


"""

from .utils import *
from .interfaces import *
from .motion_planners import *

# TODO: export version author etc.
# __all__ = [name for name in dir() if not name.startswith('_')]
